#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>
#define MODULO_UINT ((unsigned int) 0x7FFFFFFF)
#define DEBUG 1
#define ELMNTS  8
#define N_SORT_ALGS 2

typedef int       data;

void bubble(void *, unsigned int , unsigned int , int (*compare) (const void * , const void *));


int compare(const void *x,const void *y)
{
  data  xp, yp;

  xp = *((data *) x);
  yp = *((data *) y);
  if (xp>yp) return(1);
  if (xp == yp) return(0);
  return(-1);
}

static struct timeval _tstart, _tend;
static struct timezone _tstruct;

void   avg_tstart  (void);
void   avg_tend    (void);
double avg_treturn (void);

void avg_tstart(void) {
  gettimeofday(&_tstart, &_tstruct);
}

void avg_tend(void) {
  gettimeofday(&_tend,&_tstruct);
}

double avg_treturn(void) {
  double t1,t2;
  t1=(double)_tstart.tv_sec + (double)_tstart.tv_usec/(1000*1000);
  t2=(double)_tend.tv_sec + (double)_tend.tv_usec/(1000*1000);
  return(t2-t1);
}

int main(int argc,char *argv[]) {
  int i,j,k,n,runs, type;
  data *ptr,*bptr;
  double elapsed;
  void (*gfunc[N_SORT_ALGS])(void *,unsigned int   ,unsigned int,int (*) (const void *,const void *)) = 
                                {  qsort,	  bubble  };
  char *labels[N_SORT_ALGS]    = { "qsort",      "bubble" };
  int doalg[N_SORT_ALGS]       = { 1,1 };

  void (*GTYPESORT)(void *,unsigned int,unsigned int,int (*) (const void *,const void *));

  n=1000;runs=1;
  if (( argc <= 1 ) || (argc > 4)) {
    printf("Usage: %s size input-distrib(from 1 to 5)\n",argv[0]);
    return(1);
  } 
  else {
    n = atoi(argv[1]);
    type=atoi(argv[2]);
   }

   ptr = (data*) malloc(n*sizeof(data));
   bptr= (data*) malloc(n*sizeof(data));
   memset( ptr,0,n*sizeof(data));
   memset(bptr,0,n*sizeof(data));

   srandom(21+1001*11);
   for(i=0;i<n;i++)  {
    if  ( type==1 ) {
     bptr[i]=ptr[i]=(data) ((n)*(random()/((double)INT_MAX)));
    }
    if  ( type==2 ) {
     bptr[i]=ptr[i]=(data) 10;
    }
    if  ( type==3 ) {
     bptr[i]=ptr[i]=(data) i+1;
    }
    if  ( type==4 ) {
     bptr[i]=ptr[i]=(data) n-i;
    }
    if  ( type>=5 ) {
     bptr[i]=ptr[i]=(data) (random()&MODULO_UINT);
    }
   }

   for(k=0;k<N_SORT_ALGS;k++) {
    if (doalg[k]==1) {
     GTYPESORT=gfunc[k];

     for(i=0;i<n;i++) ptr[i] = bptr[i];
     if (DEBUG) { 
        for(i=0;i<ELMNTS;i++) printf("%d ",(int)ptr[i]); printf("..."); 
        for(i=n-ELMNTS;i<n;i++) printf("%d ",(int)ptr[i]); printf("\n");
     }
     elapsed=0.0;  avg_tstart();

     GTYPESORT(ptr,n,sizeof(data),compare);

     avg_tend(); elapsed += (avg_treturn());
     if (DEBUG){
        for(i=0;i<ELMNTS;i++) printf("%d ",(int)ptr[i]); printf("..."); 
        for(i=n-ELMNTS;i<n;i++) printf("%d ",(int)ptr[i]); printf("\n");
     }
     for(i=0;i<n-1;i++) if (ptr[i]>ptr[i+1]) {printf("Not Sorted\n"); }
     printf("%12s : Elapsed time is: %10.8f\n",labels[k],elapsed/runs);
    }
   }

   free((void *) bptr);
   free((void *) ptr);

   return(0);
}
